import torch
from scipy.special import gamma, gammainc, gammaincinv

'''
Helper functions for converting points from a gaussian distribution to a uniform distribution
on the unit sphere and vice versa.
'''

def gaussian_to_sphere(X: torch.Tensor) -> torch.Tensor:
    # Convert a multivariate gaussian of any dimension d to a d-dimensional sphere
    d = X.shape[1]
    # Compute the norm of each row
    norm = torch.norm(X, dim=1, keepdim=True)

    # compute cdf of each point
    cdf = gammainc(d/2, norm**2/2) ** (1/d)

    # calculate the new point with the adjusted radius
    X_sphere = X / norm * cdf
    return X_sphere

def sphere_to_gaussian(X: torch.Tensor) -> torch.Tensor:
    # Convert a d-dimensional sphere to a multivariate gaussian of any dimension d
    d = X.shape[1]
    # Compute the norm of each row
    norm = torch.norm(X, dim=1, keepdim=True)

    # compute cdf of each point
    inv_cdf = (gammaincinv(d/2, (norm ** d))*2)**0.5

    # calculate the new point with the adjusted radius
    X_gaussian = X / norm * inv_cdf
    return X_gaussian



import torch
import torch.nn as nn
import torch.optim as optim
import numpy as np

from torch.utils.data import DataLoader
from torchdiffeq import odeint
from tqdm import tqdm
from scipy.stats import shapiro, combine_pvalues

from pingouin import multivariate_normality

from LAMINAR.utils.gaussian2uniform import gaussian_to_sphere, sphere_to_gaussian

'''
This code was partly taken from Ricky Chen's implementation of the Neural ODEs paper* and modified for our purposes.

GitHub: https://github.com/rtqichen/torchdiffeq/blob/master/examples/cnf.py

*Chen, R. T. Q., Rubanova, Y., Bettencourt, J., & Duvenaud, D. (2019). Neural Ordinary Differential Equations. arXiv. https://arxiv.org/abs/1806.07366.
'''


# define the hypernetwork, generating the time dependent parameters of the planar CNF
class HyperNetwork(nn.Module):
    def __init__(self,
                 in_out_dim: int = 2,
                 hidden_dim: int = 32,
                 width: int = 64,
                 device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
                 ):
        '''
        in_out_dim: int         - dimensionality of the input and output space
        hidden_dim: int         - dimensionality of the hidden layer of the MLP
        width: int              - width of the planar CNF
        device: torch.device    - device on which the model is trained, either GPU if available or CPU
        '''

        super(HyperNetwork, self).__init__()

        self.device = device

        self.blocksize = in_out_dim * width
        self.in_out_dim = in_out_dim
        self.width = width

        # fully conencted MLP with one hidden layer
        self.fc1 = nn.Linear(1, hidden_dim).to(self.device)
        self.fc2 = nn.Linear(hidden_dim, hidden_dim).to(self.device)
        self.fc3 = nn.Linear(hidden_dim, 3 * self.blocksize + width).to(self.device)

    def forward(self,
                 t: torch.Tensor) -> list:
        '''
        t: torch.Tensor - time parameter for the forward pass
        '''

        # forward pass through the MLP
        parameters = t.reshape(1, 1)
        parameters = torch.tanh(self.fc1(parameters))
        parameters = torch.tanh(self.fc2(parameters))
        parameters = self.fc3(parameters)

        # split the output into the parameters of the planar CNF
        parameters = parameters.reshape(-1)
        W = parameters[:self.blocksize].reshape(self.width, self.in_out_dim, 1).to(self.device)

        U = parameters[self.blocksize:2 * self.blocksize].reshape(self.width, 1, self.in_out_dim).to(self.device)
        G = parameters[2 * self.blocksize:3 * self.blocksize].reshape(self.width, 1, self.in_out_dim).to(self.device)
        U = U * torch.sigmoid(G)

        B = parameters[3 * self.blocksize:].reshape(self.width, 1, 1).to(self.device)

        return [W, B, U]


# function for calculating the trace of the jacobian w.r.t. z
def trace_df_dz(f: nn.Module,
                z: torch.Tensor) -> torch.Tensor:
    '''
    f: nn.Module    - function for which the trace of the jacobian w.r.t. z is calculated
    z: torch.Tensor - input tensor for the function f
    '''
    sum_diag = 0
    for i in range(z.shape[1]):
       sum_diag += torch.autograd.grad(f[:, i].sum(), z, create_graph=True)[0].contiguous()[:, i].contiguous()
    return sum_diag


# actual planar flow model
class PlanarCNF(nn.Module):
    def __init__(self,
                 in_out_dim: int = 2,
                 hidden_dim: int = 32,
                 width: int = 64,
                 device=torch.device("cuda" if torch.cuda.is_available() else "cpu")):
        '''
        in_out_dim: int         - dimensionality of the input and output space
        hidden_dim: int         - dimensionality of the hidden layer of the HyperNetwork
        width: int              - width of the planar CNF
        device: torch.device    - device on which the model is trained, either GPU if available or CPU
        '''    
    
        super(PlanarCNF, self).__init__()
        self.device = device
        self.hypernetwork = HyperNetwork(in_out_dim, hidden_dim, width, device)

    # forward pass of the planar CNF
    def forward(self,
                t: torch.Tensor,
                states: torch.Tensor) -> tuple:
        '''
        t: torch.Tensor         - time parameter for the forward pass
        states: torch.Tensor    - input tensor for the forward pass, containing the states z and the log probability of z
        '''

        z = states[0]
        logp_z = states[1]

        batch_size = z.shape[0]

        with torch.set_grad_enabled(True):
            z.requires_grad_(True)

            W, B, U = self.hypernetwork(t)

            Z = torch.unsqueeze(z, 0).repeat(W.shape[0], 1, 1)

            h = torch.tanh(torch.matmul(Z, W) + B)
            dz_dt = torch.matmul(h, U).mean(0)

            dlogp_z_dt = -trace_df_dz(dz_dt, z).view(batch_size, 1)

        return (dz_dt, dlogp_z_dt)

    # loss function for the planar CNF
    def loss(self,
             z_t0: torch.Tensor,
             logp_diff_t0: torch.Tensor) -> torch.Tensor:
        '''
        z_t0: torch.Tensor          - initial state of the planar CNF
        logp_diff_t0: torch.Tensor  - log probability of the initial state
        '''        
    
        p_z0 = torch.distributions.MultivariateNormal(
            loc=torch.zeros(z_t0.shape[1]).to(self.device),
            covariance_matrix=torch.eye(z_t0.shape[1]).to(self.device))

        logp_x = p_z0.log_prob(z_t0) - logp_diff_t0.view(-1)
        return -logp_x.mean(0)

    # transform data to the latent space and vice versa
    # takes care of time steps and the transformation direction
    def transform(self,
                  data: torch.Tensor,
                  timesteps: int = 50,
                  reverse: bool = False) -> torch.Tensor:
        '''
        data: torch.Tensor  - input data to be transformed
        timesteps: int      - number of time steps for the transformation
        reverse: bool       - direction of the transformation, either data->latent space or latent space->data
        '''
        self.eval()

        if not reverse:  # transform data to gaussian -> sphere
            with torch.no_grad():
                z_t, _ = odeint(
                    self,
                    (data, torch.zeros(data.shape[0], 1).to(self.device)),
                    torch.linspace(1, 0, timesteps).type(torch.float32).to(self.device),
                    atol=1e-5, rtol=1e-5,
                    method='dopri5'
                )

                z_t = gaussian_to_sphere(z_t[-1].detach().cpu()).to(self.device)

        else:  # transform sphere->gaussian to data space
            with torch.no_grad():
                z_t, _ = odeint(
                    self,
                    (sphere_to_gaussian(data.detach().cpu()).to(self.device), torch.zeros(data.shape[0], 1).to(self.device)),
                    torch.linspace(0, 1, timesteps).type(torch.float32).to(self.device),
                    atol=1e-5, rtol=1e-5,
                    method='dopri5'
                )

                z_t = z_t[-1]
        return z_t


#class EarlyStopping:
#    def __init__(self, 
#                 patience: int = 15, 
#                 significance: float = 5):
#        
#        self.patience = patience
#        self.significance = significance
#
#    def __call__(self, loss_hist: list) -> bool:
#        if len(loss_hist) <= self.patience:
#            return False
#        
#        vals = torch.tensor(loss_hist[-self.patience-1:-1])
#        vals_mean = vals.mean()
#        vals_std = vals.std()
#
#        if loss_hist[-1] > vals_mean + self.significance * vals_std:
#            return True

class EarlyStopping:
    def __init__(self, patience: int = 5, p_lim: float = 0.05):
        self.patience = patience
        self.p_lim= p_lim

        self.p_hist = []

    def __call__(self, p_val) -> bool:
        self.p_hist.append(p_val)

        if len(self.p_hist) <= self.patience:
            return False
        
        if (torch.tensor(self.p_hist[-self.patience:]) > self.p_lim).all():
            return True

# train function of the planar CNF    
def train_PlanarCNF(
        model: PlanarCNF,
        optimizer: optim.Optimizer,
        train_loader: DataLoader,
        epochs: int = 100,
        batch_size: int = 128,
        patience: int = 15,  # early stopping patience
        p_lim: float = 0.05, # early stopping significance
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu"),
        verbose: bool = True) -> list:
    '''
    model: PlanarCNF                - model to be trained
    optimizer: optim.Optimizer      - optimizer for the training process
    train_loader: DataLoader        - data loader for the training data
    epochs: int                     - number of epochs for the training
    batch_size: int                 - batch size for the training
    patience: int                   - early stopping patience
    device: torch.device            - device on which the model is trained, either GPU if available or CPU
    verbose: bool                   - verbosity of the training process
    '''
    
    bad_p_val = False

    loss_history = []

    model.to(device)
    model.train()
    dataloader_train = DataLoader(train_loader, batch_size=batch_size, shuffle=True)

    pbar = tqdm(range(epochs), disable=not verbose)

    earlystop = EarlyStopping(patience=patience, p_lim=p_lim)
    
    for epoch in pbar:
        total_loss = 0
        for i, batch in enumerate(dataloader_train):
            batch = batch.to(device)
            logp_diff_t1 = torch.zeros(batch.shape[0], 1).to(device)

            optimizer.zero_grad()

            # solve ode
            z_t, logp_diff_t = odeint(
                model, 
                (batch, logp_diff_t1),
                torch.tensor([1, 0]).type(torch.float32).to(device),
                atol=1e-5, rtol=1e-5,
                method='dopri5',
            )

            z_t0, logp_diff_t0 = z_t[-1], logp_diff_t[-1]

            loss = model.loss(z_t0, logp_diff_t0)
            loss.backward()
            optimizer.step()
            total_loss += loss.item()

        loss_history.append(total_loss / len(dataloader_train))

        if not bad_p_val:
            pushed = sphere_to_gaussian(model.transform(train_loader, reverse=False).detach().cpu())
            try:
                p_value = multivariate_normality(pushed.cpu().detach().numpy())[1]
            except np.linalg.LinAlgError:
                print('Unable to calculate p-value - Deactive early stopping with p-value monitoring')
                p_value = 0.0
                bad_p_val = True

        if verbose:
            pbar.set_description(f"Epoch {epoch+1} | Loss: {loss_history[-1]:.4f} | p-value: {p_value:.2E}")

        #if p_value > 0.05:
        #    if verbose:
        #        print(f"Early stopping at epoch {epoch+1}")
        #    break

        # early stopping
        if not bad_p_val:
            if earlystop(p_value):
                if verbose:
                    print(f"Early stopping at epoch {epoch+1}")
                break

    return loss_history



import torch
import numpy as np

from typing import Union
from scipy.spatial import KDTree
from scipy.sparse.csgraph import dijkstra
from scipy.stats import shapiro, combine_pvalues    
from pingouin import multivariate_normality
from LAMINAR.Flow.planarCNF import PlanarCNF, train_PlanarCNF
from LAMINAR.utils.gaussian2uniform import sphere_to_gaussian


'''
Implementation of the LAM algorithm using a normalizing flow to transform the data
'''
class LAMINAR():
    def __init__(self,
                 data: Union[np.ndarray, torch.Tensor],
                 epochs: int = 100,
                 k_neighbours: int = 20,
                 hyperparameters: dict = {},
                 device = torch.device("cuda" if torch.cuda.is_available() else "cpu")):
        '''
        data: Union[np.ndarray, torch.Tensor]  - input data to be transformed, either array or tensor
        epochs: int                            - number of (max) epochs for the training, early stopping is used
        k_neighbours: int                      - number of neighbours to consider; more neighbours lead to a more global metric but also to a higher computational cost, less neighbours lead to more local metric
        hyperparameters: dict                  - hyperparameters for the flow model
        device: torch.device                   - device on which the model is trained, either GPU if available or CPU
        '''

        self.device = device   
        self.k_neighbours = k_neighbours
        self.dimension = data.shape[1]

        # make data a tensor if it is an array
        if isinstance(data, np.ndarray):
            data = torch.tensor(data, dtype=torch.float32)

        self.data = data.to(self.device)

        # hyperparameters for the flow model
        self.hidden_dim = hyperparameters.get('hidden_dim', 32)
        self.width = hyperparameters.get('width', 64)
        self.timesteps = hyperparameters.get('timesteps', 50)
        self.learning_rate = hyperparameters.get('learning_rate', 1e-3)
        self.patience = hyperparameters.get('patience', 5)
        self.p_lim = hyperparameters.get('p_lim', 5.0)
        self.batch_size = hyperparameters.get('batch_size', 128)

        # initialize the flow
        self.flow = PlanarCNF(in_out_dim=self.dimension, hidden_dim=self.hidden_dim, width=self.width, device=self.device)
        optimizer = torch.optim.Adam(self.flow.parameters(), lr=self.learning_rate)
        
        # train the flow on data
        self._train(self.data, optimizer, epochs=epochs, batch_size=self.batch_size, patience=self.patience, p_lim=self.p_lim)
        
        # push the data throught the flow
        self.data_pushed = self.flow.transform(self.data, timesteps=self.timesteps)
        
        # generate the distance matrix of the k closest neighbourhoods
        self._generate_distance_matrix()

    # get p value of the gaussian after the flow
    def p_value(self):
        '''
        Function to calculate the p-value of the pushed data distribution
        '''
        # calculate the p-value of the data distribution

        data = sphere_to_gaussian(self.data_pushed.cpu().detach()).to(self.device)
        try:
            p = multivariate_normality(data.cpu().detach().numpy())[1]
        except np.linalg.LinAlgError:
            print('Unable to calculate p-value')
            p = 0.0

        print(f'Henze-Zirkler p-value:\t{p}')
    
        return p 

    # train function
    def _train(self,
              data: torch.Tensor,
              optimizer: torch.optim.Optimizer,
              epochs: int = 100,
              batch_size: int = 128,
              patience: int = 5,
              p_lim: float = 0.05,
              verbose: bool = True
              ):
        '''
        data: torch.Tensor                  - input data to be transformed
        optimizer: torch.optim.Optimizer    - optimizer for the training process
        epochs: int                         - number of epochs for the training
        batch_size: int                     - batch size for the training
        patience: int                       - early stopping patience
        p_lim: float                        - p_lim level for the early stopping
        verbose: bool                       - verbosity of the training process
        '''
        self.loss_history = train_PlanarCNF(self.flow, optimizer, data, epochs, batch_size, patience, p_lim, self.device, verbose)    

    # function to generate the distance matrix for the neighbourhoods
    def _generate_distance_matrix(self):
        # get the neighbours of each point
        indices = self._internal_query(range(self.data.shape[0]))
        
        # calculate the covariance of each points neighbours
        cov_matrices = np.zeros((self.data.shape[0], self.dimension, self.dimension))
        for i in range(self.data.shape[0]):
            cov_matrices[i] = np.cov(self.data[indices[i]].cpu().detach().numpy().T)

        self.cov_matrices = torch.tensor(cov_matrices, dtype=torch.float32).to(self.device)

        # initialize the matrix with infinity
        self.distance_matrix = torch.zeros((self.data.shape[0], self.data.shape[0])).to(self.device)
        self.distance_matrix.fill_(float('inf'))

        for i in range(self.data.shape[0]):
            for j in indices[i]:
                if self.distance_matrix[i, j] == float('inf'):
                    # calculate the mahalanobis distance between the points
                    common_cov = (self.cov_matrices[i] + self.cov_matrices[j])/2 + 1e-6*torch.eye(self.dimension).to(self.device)

                    x_i = self.data[i].reshape(1, -1)
                    x_j = self.data[j].reshape(1, -1)

                    # determinant to keep the rescaling in mind
                    cov_det = torch.det(common_cov) ** 1/self.dimension

                    mahalanobis_distance = torch.sqrt(cov_det * (x_i - x_j) @ torch.inverse(common_cov) @ (x_i - x_j).T)

                    self.distance_matrix[i, j] = mahalanobis_distance
                    self.distance_matrix[j, i] = mahalanobis_distance

    # internal query for the initial k nearest neighbours
    def _internal_query(self,
                        indices: Union[int, np.ndarray],   # indices of the points to query
                        k_neighbours: Union[int, None] = None) -> np.ndarray:
        '''
        indices: np.ndarray            - indices of the points to query
        k_neighbours: Union[int, None] - number of neighbours to consider; None means the default value of the class is used
        '''

        if isinstance(indices, int):
            indices = [indices]
        
        if k_neighbours is None:
            k_neighbours = self.k_neighbours

        points = self.data[indices]

        points = self.flow.transform(points)
        
        # use kdtrees for the neighourhood search        
        kdtree = KDTree(self.data_pushed.cpu().detach().numpy())
        _, indices = kdtree.query(points.cpu().detach().numpy(), k=k_neighbours)

        return indices
    
    # query for the user to get the k nearest neighbours and their distances
    def query(self,
              indices: Union[int, np.ndarray],   # indices of the points to query
              k_neighbours: Union[int, None] = None) -> np.ndarray:
        '''
        indices: np.ndarray            - indices of the points to query
        k_neighbours: Union[int, None] - number of neighbours to consider; None means the default value of the class is used
        '''
        if isinstance(indices, int):
            indices = [indices]

        if k_neighbours is None:
            k_neighbours = self.k_neighbours

        dist = self.distance(indices, return_path=False)
        ind = []
        d = []

        for i in range(len(indices)):
            # sort ind and dist by the distance
            ind.append(np.argsort(dist[i])[:k_neighbours])
            d.append(np.sort(dist[i])[:k_neighbours])

        return ind, d   
    
    # function to calculate the distance between any points
    def distance(self, 
                 x_ind: np.ndarray,
                 y_ind: Union[np.ndarray, None] = None,
                 return_path: bool = False) -> float:
        '''
        x_ind: np.ndarray               - index of the reference point
        y_ind: Union[np.ndarray, None]  - index of the target point(s); None selects all points
        return_path: bool               - return all points on the shortest path
        '''
        dist, path = dijkstra(self.distance_matrix.cpu().detach().numpy(), indices=x_ind, return_predecessors=True)
        if y_ind is None:
            return dist
        
        else:
            dist = dist[y_ind]

            # path reconstruction only if a single target point is given
            if (return_path and isinstance(y_ind, int)) or (return_path and isinstance(y_ind, np.ndarray) and len(y_ind) == 1):
                shortest_path = [y_ind]
                while shortest_path[-1] != x_ind:
                    shortest_path.append(path[shortest_path[-1]])
                return dist, shortest_path[::-1]
            
            else:
                return dist


#from .Flow import *
from .LAMINAR import LAMINAR
from .utils import standardize


def add_one(x: int):
    """An example function that increases a number

    :param x: The input parameter to increase
    :return: The successor of the given number
    """
    return x + 1



import LAMINAR
import torch
import numpy as np

import LAMINAR.Flow


def test_LAMINAR():
    assert LAMINAR.add_one(1) == 2
    
    data = torch.rand(100, 2)
    LAM = LAMINAR.LAMINAR(data)

    # check basic structures
    assert LAM.data.shape == data.shape
    assert LAM.data_pushed.shape == data.shape
    assert LAM.k_neighbours == 20
    assert LAM.dimension == 2
    assert LAM.flow.device == torch.device("cuda" if torch.cuda.is_available() else "cpu")
    assert type(LAM.flow).__name__ == LAMINAR.Flow.planarCNF.PlanarCNF.__name__
    assert LAM.distance_matrix.shape == (100, 100)
    assert LAM.cov_matrices.shape == (100, 2, 2)

    # check metric properties:
    
    # build full distance matrix
    dist_matrix = np.zeros((100, 100))
    for i in range(100):
        dist = LAM.distance(i)
        dist_matrix[i, :] = dist

    # symmetry
    assert (dist_matrix == dist_matrix.T).all()

    # positive definiteness
    assert (np.diag(dist_matrix) == 0).all()
    assert (dist_matrix >= 0).all()

    # triangle inequality
    for i in range(100):
        for j in range(100):
            for k in range(100):
                assert dist_matrix[i, j] <= dist_matrix[i, k] + dist_matrix[k, j]

    
    # check standardisation
    assert LAMINAR.standardize(data).shape == data.shape
    assert LAMINAR.standardize(data).mean() <= 1e-3
    assert abs(LAMINAR.standardize(data).std() - 1) <= 1e-3
    